@@ -1105,7 +1105,7 @@ static dav_error * dav_fs_deliver(const dav_resource *resource,
     APR_BRIGADE_INSERT_TAIL(bb, bkt);
 
     if ((status = ap_pass_brigade(output, bb)) != APR_SUCCESS) {
-        return dav_new_error(pool, HTTP_FORBIDDEN, 0, status,
+        return dav_new_error(pool, AP_FILTER_ERROR, 0, status,
                              "Could not write contents to filter.");
     }
 
@@ -1663,7 +1663,7 @@ static dav_error * dav_fs_walker(dav_fs_walker_context *fsctx, int depth)
         /* put a slash back on the end of the directory */
         fsctx->path1.buf[fsctx->path1.cur_len - 1] = '/';
 
-        /* these are all non-existant (files) */
+        /* these are all non-existent (files) */
         fsctx->res1.exists = 0;
         fsctx->res1.collection = 0;
         memset(&fsctx->info1.finfo, 0, sizeof(fsctx->info1.finfo));
